/*
** File: admxrc3.c  
** Project: ADMXRC3 API module library
** Purpose: OS-independent code for ADMXRC3 API module library
**
** (C) Copyright Alpha Data 2009-2011
*/

#include <admxrc3.h>
#include <admxrc3_platform.h>
#include <portable_file.h>
#include <portable_string.h>
#if PORTABLE_HAS_WCHAR_SUPPORT
# include <portable_wstring.h>
#endif
#include "ioctl_admxrc3.h"

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

static const double g_64k = 65536.0;

#define TO_DOUBLE32(i, f) ((double)(i) + (double)(f) / g_64k);

#if 0
/* We might make this extern one day */
static ADMXRC3_STATUS
ADMXRC3_GetCooperativeLevel(
  ADMXRC3_HANDLE hCard,
  uint32_t* pCooperativeLevel);
#endif

/* We might make this extern one day */
static ADMXRC3_STATUS
ADMXRC3_SetCooperativeLevel(
  ADMXRC3_HANDLE hCard,
  uint32_t cooperativeLevel);

static const char* fpgaDeviceNamesA[] = {
  "xc4085xl",   /*  0 */
  "xc40150xv",	/*  1 */
  "xc40200xv",	/*  2 */
  "xc40250xv",	/*  3 */
  "v1000",      /*  4 */
  "v400",       /*  5 */
  "v600",       /*  6 */
  "v800",       /*  7 */
  "v2000e",     /*  8 */
  "v1000e",     /*  9 */
  "v1600e",     /* 10 */
  "v3200e",     /* 11 */
  "v812e",      /* 12 */
  "v405e",      /* 13 */
  "unknown14",	/* 14 */
  "unknown15",	/* 15 */
  "unknown16",	/* 16 */
  "unknown17",	/* 17 */
  "unknown18",	/* 18 */
  "unknown19",	/* 19 */
  "unknown20",	/* 20 */
  "unknown21",	/* 21 */
  "unknown22",	/* 22 */
  "unknown23",	/* 23 */
  "unknown24",	/* 24 */
  "unknown25",	/* 25 */
  "unknown26",	/* 26 */
  "unknown27",	/* 27 */
  "unknown28",	/* 28 */
  "unknown29",	/* 29 */
  "unknown30",	/* 30 */
  "unknown31",	/* 31 */
  "2v1000",     /* 32 */
  "2v1500",     /* 33 */
  "2v2000",     /* 34 */
  "2v3000",     /* 35 */
  "2v4000",     /* 36 */
  "2v6000",     /* 37 */
  "2v8000",     /* 38 */
  "2v10000",    /* 39 */
  "unknown40",	/* 40 */
  "unknown41",	/* 41 */
  "unknown42",	/* 42 */
  "unknown43",	/* 43 */
  "unknown44",	/* 44 */
  "unknown45",	/* 45 */
  "unknown46",	/* 46 */
  "unknown47",	/* 47 */
  "unknown48",	/* 48 */
  "unknown49",	/* 49 */
  "unknown50",	/* 50 */
  "unknown51",	/* 51 */
  "unknown52",	/* 52 */
  "unknown53",	/* 53 */
  "unknown54",	/* 54 */
  "unknown55",	/* 55 */
  "unknown56",	/* 56 */
  "unknown57",	/* 57 */
  "unknown58",	/* 58 */
  "unknown59",	/* 59 */
  "unknown60",	/* 60 */
  "unknown61",	/* 61 */
  "unknown62",	/* 62 */
  "unknown63",	/* 63 */
  "2vp2",       /* 64 */
  "2vp4",       /* 65 */
  "2vp7",       /* 66 */
  "2vp20",      /* 67 */
  "2vp30",      /* 68 */
  "2vp40",      /* 69 */
  "2vp50",      /* 70 */
  "2vp100",     /* 71 */
  "2vp125",     /* 72 */
  "2vp70",      /* 73 */
  "unknown74",	/* 74 */
  "unknown75",	/* 75 */
  "unknown76",	/* 76 */
  "unknown77",	/* 77 */
  "unknown78",	/* 78 */
  "unknown79",	/* 79 */
  "unknown80",	/* 80 */
  "unknown81",	/* 81 */
  "unknown82",	/* 82 */
  "unknown83",	/* 83 */
  "unknown84",	/* 84 */
  "unknown85",	/* 85 */
  "unknown86",	/* 86 */
  "unknown87",	/* 87 */
  "unknown88",	/* 88 */
  "unknown89",	/* 89 */
  "unknown90",	/* 90 */
  "unknown91",	/* 91 */
  "unknown92",	/* 92 */
  "unknown93",	/* 93 */
  "unknown94",	/* 94 */
  "unknown95",	/* 95 */
  "4vlx15",     /* 96 */
  "4vlx25",     /* 97 */
  "4vlx40",     /* 98 */
  "4vlx60",     /* 99 */
  "4vlx100",    /* 100 */
  "4vlx160",    /* 101 */
  "4vlx200",    /* 102 */
  "4vlx80",     /* 103 */
  "4vsx25",     /* 104 */
  "4vsx35",     /* 105 */
  "4vsx55",     /* 106 */
  "unknown106",	/* 107 */
  "unknown108",	/* 108 */
  "unknown109",	/* 109 */
  "unknown110",	/* 110 */
  "unknown111",	/* 111 */
  "4vfx12",     /* 112 */
  "4vfx20",     /* 113 */
  "4vfx40",     /* 114 */
  "4vfx60",     /* 115 */
  "4vfx100",    /* 116 */
  "4vfx140",    /* 117 */
  "unknown118",	/* 118 */
  "unknown119",	/* 119 */
  "unknown120",	/* 120 */
  "unknown121",	/* 121 */
  "unknown122",	/* 122 */
  "unknown123",	/* 123 */
  "unknown124",	/* 124 */
  "unknown125",	/* 125 */
  "unknown126",	/* 126 */
  "unknown127",	/* 127 */
  "5vlx30",     /* 128 */
  "5vlx50",     /* 129 */
  "5vlx85",     /* 130 */
  "5vlx110",    /* 131 */
  "5vlx220",    /* 132 */
  "5vlx330",    /* 133 */
  "unknown134",	/* 134 */
  "unknown135", /* 135 */
  "5vlx30t",    /* 136 */
  "5vlx50t",    /* 137 */
  "5vlx85t",    /* 138 */
  "5vlx110t",   /* 139 */
  "5vlx330t",   /* 140 */
  "5vlx220t",   /* 141 */
  "5vlx155t",   /* 142 */
  "unknown143",	/* 143 */
  "5vsx35t",    /* 144 */
  "5vsx50t",    /* 145 */
  "5vsx95t",    /* 146 */
  "5vsx240t",   /* 147 */
  "unknown148", /* 148 */
  "unknown149", /* 149 */
  "unknown150", /* 150 */
  "unknown151", /* 151 */
  "5vfx100t",   /* 152 */
  "5vfx130t",   /* 153 */
  "5vfx200t",   /* 154 */
  "5vfx30t",    /* 155 */
  "5vfx70t",    /* 156 */
  "unknown157", /* 157 */
  "unknown158", /* 158 */
  "unknown159", /* 159 */
  "6vlx760",    /* 160 */
  "unknown161", /* 161 */
  "unknown162", /* 162 */
  "unknown163", /* 163 */
  "unknown164", /* 164 */
  "unknown165", /* 165 */
  "unknown166", /* 166 */
  "unknown167", /* 167 */
  "6vlx75t",    /* 168 */
  "6vlx130t",   /* 169 */
  "6vlx195t",   /* 170 */
  "6vlx240t",   /* 171 */
  "6vlx365t",   /* 172 */
  "6vlx550t",   /* 173 */
  "unknown174", /* 174 */
  "unknown175", /* 175 */
  "6vsx315t",   /* 176 */
  "6vsx475t",   /* 177 */
  "unknown178", /* 178 */
  "unknown179", /* 179 */
  "unknown180", /* 180 */
  "unknown181", /* 181 */
  "unknown182", /* 182 */
  "unknown183", /* 183 */
  "6vhx250t",   /* 184 */
  "6vhx255t",   /* 185 */
  "6vhx380t",   /* 186 */
  "6vhx565t",   /* 187 */
	"unknown188", /* 188 */
	"unknown189", /* 189 */
	"unknown190", /* 190 */
	"unknown191", /* 191 */
	"unknown192", /* 192 */
	"unknown193", /* 193 */
	"unknown194", /* 194 */
	"unknown195", /* 195 */
	"unknown196", /* 196 */
	"unknown197", /* 197 */
	"unknown198", /* 198 */
	"unknown199", /* 199 */
	"unknown200", /* 200 */
	"unknown201", /* 201 */
	"unknown202", /* 202 */
	"unknown203", /* 203 */
	"unknown204", /* 204 */
	"unknown205", /* 205 */
	"unknown206", /* 206 */
	"unknown207", /* 207 */
	"unknown208", /* 208 */
	"unknown209", /* 209 */
	"unknown210", /* 210 */
	"unknown211", /* 211 */
	"unknown212", /* 212 */
	"unknown213", /* 213 */
	"unknown214", /* 214 */
	"unknown215", /* 215 */
	"unknown216", /* 216 */
	"unknown217", /* 217 */
	"unknown218", /* 218 */
	"unknown219", /* 219 */
	"unknown220", /* 220 */
	"unknown221", /* 221 */
	"unknown222", /* 222 */
	"unknown223", /* 223 */
  "7a100t",     /* 224 */
  "7a200t",     /* 225 */
  "7a350t",     /* 226 */
	"unknown227", /* 227 */
	"unknown228", /* 228 */
	"unknown229", /* 229 */
	"unknown230", /* 230 */
	"unknown231", /* 231 */
  "7k70t",      /* 232 */
  "7k160t",     /* 233 */
  "7k325t",     /* 234 */
	"7k355t",     /* 235 */
	"7k410t",     /* 236 */
	"7k420t",     /* 237 */
	"7k480t",     /* 238 */
	"unknown239", /* 239 */
	"unknown240", /* 240 */
	"unknown241", /* 241 */
	"unknown242", /* 242 */
	"unknown243", /* 243 */
	"unknown244", /* 244 */
	"unknown245", /* 245 */
	"unknown246", /* 246 */
	"unknown247", /* 247 */
  "7v585t",     /* 248 */
  "7v1500t",    /* 249 */
  "7v2000t",    /* 250 */
  "unknown251", /* 251 */
  "unknown252", /* 252 */
  "unknown253", /* 253 */
  "unknown254", /* 254 */
  "unknown255", /* 255 */
	"7vx330t",    /* 256 */
	"7vx415t",    /* 257 */
	"7vx485t",    /* 258 */
	"7vx550t",    /* 259 */
	"7vx690t",    /* 260 */
	"7vx980t",    /* 261 */
	"7vx1140t",   /* 262 */
  "unknown263", /* 263 */
  "unknown264", /* 264 */
  "unknown265", /* 265 */
  "unknown266", /* 266 */
  "unknown267", /* 267 */
  "unknown268", /* 268 */
  "unknown269", /* 269 */
  "unknown270", /* 270 */
  "unknown271", /* 271 */
	"7vh290t",    /* 272 */
	"7vh580t",    /* 273 */
	"7vh870t",    /* 274 */
	"unknown275", /* 275 */
	"unknown276", /* 276 */
	"unknown277", /* 277 */
	"unknown278", /* 278 */
  "unknown279", /* 279 */
  "unknown280", /* 280 */
  "unknown281", /* 281 */
  "unknown282", /* 282 */
  "unknown283", /* 283 */
  "unknown284", /* 284 */
  "unknown285", /* 285 */
  "unknown286", /* 286 */
  "unknown287", /* 287 */
  "7z010",      /* 288 */
  "7z015",      /* 289 */
  "7z020",      /* 290 */
  "7z030",      /* 291 */
  "7z045",      /* 292 */
  "7z100",      /* 293 */
  "unknown294", /* 294 */
  "unknown295", /* 295 */
  "unknown296", /* 296 */
  "unknown297", /* 297 */
  "unknown298", /* 298 */
  "unknown299", /* 299 */
  "unknown300", /* 300 */
  "unknown301", /* 301 */
  "unknown302", /* 302 */
  "unknown303", /* 303 */
  "unknown304", /* 304 */
  "unknown305", /* 305 */
  "unknown306", /* 306 */
  "unknown307", /* 307 */
  "unknown308", /* 308 */
  "unknown309", /* 309 */
  "unknown310", /* 310 */
  "unknown311", /* 311 */
  "unknown312", /* 312 */
  "unknown313", /* 313 */
  "unknown314", /* 314 */
  "unknown315", /* 315 */
  "unknown316", /* 316 */
  "unknown317", /* 317 */
  "unknown318", /* 318 */
  "unknown319", /* 319 */
  "unknown320", /* 320 */
  "ku035",      /* 321 */
  "ku040",      /* 322 */
  "ku060",      /* 323 */
  "ku075",      /* 324 */
  "ku100",      /* 325 */
  "ku115",      /* 326 */
  "unknown327", /* 327 */
  "unknown328", /* 328 */
  "unknown329", /* 329 */
  "unknown330", /* 330 */
  "unknown331", /* 331 */
  "unknown332", /* 332 */
  "vu065",      /* 333 */
  "vu080",      /* 334 */
  "vu095",      /* 335 */
  "vu125",      /* 336 */
  "vu145",      /* 337 */
  "vu160",      /* 338 */
  "vu440",      /* 339 */
  "unknown340", /* 340 */
  "unknown341", /* 341 */
  "unknown342", /* 342 */
  "unknown343", /* 343 */
  "unknown344", /* 344 */
  "unknown345", /* 345 */
  "unknown346", /* 346 */
  "unknown347"  /* 347 */
};

static ADMXRC3_STATUS
selectMapOffset(
  portable_file_t hFile,
  char* identifier,
  size_t identifierSize,
  uint64_t* pSize,
  uint64_t* pOffset)
{
  typedef enum {
    StatePreamble = 0,
    StateRecords  = 1,
    StateQuit     = 2
  } State;
  const int blocksize = 1024;
  uint8_t tmpBuf[1024];
  State state;
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  uint64_t fileSize, remaining, pos, numRead;
  bool_t bFound, bMore, bResult;
  size_t chunk, offset, level, skip;
  uint16_t recLen;
  uint8_t recType;

  if (!portableFileGetSize(hFile, &fileSize)) {
    status = ADMXRC3_INVALID_BITSTREAM;
    goto done;
  }
  
  /*
  ** Find the beginning of the SelectMap data in the file
  */
  remaining = fileSize;
  state = StatePreamble;
  bFound = FALSE;
  pos = 0;
  offset = 0;
  level = 0;
  bMore = TRUE;
  skip = 0;
  while (!bFound) {
    if (bMore) {
      if (!remaining) {
        status = ADMXRC3_INVALID_BITSTREAM;
        goto done;
      }
      memmove(tmpBuf + 0, tmpBuf + offset, level);
      offset = level;
      chunk = blocksize - level;
      if (remaining < chunk) {
        chunk = (size_t)remaining;
      }
      bResult = portableFileRead(hFile, tmpBuf, chunk, &numRead);
      if (!bResult || numRead != chunk) {
        status = ADMXRC3_INVALID_BITSTREAM;
        goto done;
      }
      level += (size_t)numRead; /* safe because numRead guaranteed small */
      bMore = FALSE;
      continue;
    } else {
      if (0 != skip) {
        chunk = level;
        if (chunk > skip) {
          chunk = skip;
        }
        level -= chunk;
        offset += chunk;
        skip -= chunk;
        pos += chunk;
        if (0 != skip) {
          bMore = TRUE;
        }
      } else {
        switch (state) {
      	case StatePreamble:
          if (level < 2) {
            bMore = TRUE;
          } else {
            /* Skip over the preamble */
            recLen = (uint16_t)(((uint16_t)tmpBuf[offset] << 8) + tmpBuf[offset + 1]);
            skip = recLen + 2;
            offset += 2;
            level -= 2;
            pos += 2;
            state = StateRecords;
          }
          break;
        case StateRecords:
          if (level < 1) {
            bMore = TRUE;
          } else {
            recType = tmpBuf[offset];
            switch (recType) {
            case 0x62:
              /*
              ** Part type record
              */
              if (level < 3) {
                bMore = TRUE;
              } else {
                recLen = (uint16_t)(((uint16_t)tmpBuf[offset + 1] << 8) + tmpBuf[offset + 2]);
                if (recLen >= identifierSize * sizeof(char)) {
                  status = ADMXRC3_INVALID_BITSTREAM;
                  goto done;
                }
                if (level < (recLen + 3U)) {
                  bMore = TRUE;
                } else {
                  memcpy(identifier, tmpBuf + offset + 3, recLen);
                  identifier[recLen] = '\0';
                  portableStringLower(identifier, identifierSize);
                  offset += recLen + 3U;
                  level -= recLen + 3U;
                  pos += recLen + 3U;
                }
              }
              break;
            case 0x65:
              /* End of header record */
              skip = 5;
              state = StateQuit;
              break;
            case 0x61:
            case 0x63:
            case 0x64:
              /* Unknown record types */
              if (level < 3) {
                bMore = TRUE;
              } else {
                recLen = (uint16_t)(((uint16_t)tmpBuf[offset + 1] << 8) + tmpBuf[offset + 2]);
                if (level < (recLen + 3U)) {
                  bMore = TRUE;
                } else {
                  offset += recLen + 3U;
                  level -= recLen + 3U;
                  pos += recLen + 3U;
                }
              }
              break;
            default:
              /* Unexpected record type - ignore it */
              recLen = (uint16_t)(((uint16_t)tmpBuf[offset + 1] << 8) + tmpBuf[offset + 2]);
              skip = recLen + 3U;
              break;
            }
          }
          break;
        case StateQuit:
          bFound = TRUE;
          break;
        }
      }
    }
  }

  *pSize = fileSize - pos;
  *pOffset = pos;

done:
  return status;
}

static ADMXRC3_STATUS
loadBitstreamA(
  portable_file_t hFile,
  ADMXRC3_BITSTREAMA** ppBitstream)
{
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  ADMXRC3_BITSTREAMA* pBitstream = NULL;
  char identifier[32];
  uint64_t size, offset, numRead;
  uint64_t totalSize;
  bool_t bResult;

  status = selectMapOffset(hFile, identifier, ARRAY_LENGTH(identifier), &size, &offset);
  if (ADMXRC3_SUCCESS != status) {
    goto done;
  }

  totalSize = sizeof(ADMXRC3_BITSTREAMA) + size; /* Will have 4 "wasted" bytes at end, but that's OK */
  if (totalSize >= 0x100000000ULL) {
    status = ADMXRC3_INVALID_BITSTREAM;
    goto done;
  }
  pBitstream = (ADMXRC3_BITSTREAMA*)malloc((size_t)totalSize); /* Cast is safe because we guarantee totalSize < 4G from above checks */
  if (NULL == pBitstream) {
    status = ADMXRC3_NO_MEMORY;
    goto done;
  }

  portableStringCopy(pBitstream->Identifier, ARRAY_LENGTH(pBitstream->Identifier), identifier);
  pBitstream->Length = (uint32_t)size; /* Cast is safe because we guarantee size < 4G from above checks */

  portableFileSetPointer(hFile, offset);
  bResult = portableFileRead(hFile, pBitstream->Data, size, &numRead);
  if (!bResult || numRead != size) {
    status = ADMXRC3_INVALID_BITSTREAM;
    goto done;
  }

  *ppBitstream = pBitstream;

done:
  if (status != ADMXRC3_SUCCESS) {
    if (pBitstream != NULL) {
      free(pBitstream);
      pBitstream = NULL;
    }
  }
  portableFileClose(hFile);
  return status;
}

#if PORTABLE_HAS_WCHAR_SUPPORT
static ADMXRC3_STATUS
loadBitstreamW(
  portable_file_t hFile,
  ADMXRC3_BITSTREAMW** ppBitstream)
{
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  ADMXRC3_BITSTREAMW* pBitstream = NULL;
  char identifier[32];
  uint64_t size, offset, numRead;
  uint64_t totalSize;
  bool_t bResult;

  status = selectMapOffset(hFile, identifier, ARRAY_LENGTH(identifier), &size, &offset);
  if (ADMXRC3_SUCCESS != status) {
    goto done;
  }

  totalSize = sizeof(ADMXRC3_BITSTREAMW) + size; /* Will have 4 "wasted" bytes at end, but that's OK */
  if (totalSize >= 0x100000000ULL) {
    status = ADMXRC3_INVALID_BITSTREAM;
    goto done;
  }
  pBitstream = (ADMXRC3_BITSTREAMW*)malloc((size_t)totalSize); /* Cast is safe because we guarantee totalSize < 4G from above checks */
  if (NULL == pBitstream) {
    status = ADMXRC3_NO_MEMORY;
    goto done;
  }

  portableUTF8ToWideChar(pBitstream->Identifier, ARRAY_LENGTH(pBitstream->Identifier), identifier);
  pBitstream->Length = (uint32_t)size; /* Cast is safe because we guarantee size < 4G from above checks */

  portableFileSetPointer(hFile, offset);
  bResult = portableFileRead(hFile, pBitstream->Data, size, &numRead);
  if (!bResult || numRead != size) {
    status = ADMXRC3_INVALID_BITSTREAM;
    goto done;
  }
  
  *ppBitstream = pBitstream;
  
done:
  if (status != ADMXRC3_SUCCESS) {
    if (pBitstream != NULL) {
      free(pBitstream);
      pBitstream = NULL;
    }
  }
  portableFileClose(hFile);
  return status;
}
#endif

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_Cancel(
  ADMXRC3_HANDLE hCard)
{
  return admxrc3IoctlCancel(hCard);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_Close(
  ADMXRC3_HANDLE hCard)
{
  /* This has to be done in an OS-specific way */
  return admxrc3CloseDevice(hCard);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_ConfigureFromBuffer(
  ADMXRC3_HANDLE hCard,
  unsigned int targetIndex,
  uint32_t flags,
  const void* pBuffer,
  size_t length)
{
  IOCTLS_ADMXRC3_CONFIGURE ioctl;
  ADMXRC3_STATUS status;

  if (length != 0 && pBuffer == NULL) {
    return ADMXRC3_NULL_POINTER;
  }

  ioctl.in.targetIndex = (uint32_t)targetIndex;
  ioctl.in.flags = flags;
  ioctl.in.pBuffer = pBuffer;
  ioctl.in.length = length;
  status = admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_CONFIGURE, &ioctl, sizeof(ioctl.in), 0);
  return status;
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_ConfigureFromFileA(
  ADMXRC3_HANDLE hCard,
  unsigned int targetIndex,
  uint32_t flags,
  const char* pFilename)
{
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  ADMXRC3_BITSTREAMA* pBitstream = NULL;
  ADMXRC3_FPGA_INFOA fpgaInfo;
  IOCTLS_ADMXRC3_CONFIGURE ioctl;
  
  status = ADMXRC3_LoadBitstreamA(pFilename, &pBitstream);
  if (ADMXRC3_SUCCESS != status) {
    goto done;
  }

  status = ADMXRC3_GetFpgaInfoA(hCard, targetIndex, &fpgaInfo);
  if (ADMXRC3_SUCCESS != status) {
    goto done;
  }

  if (portableStringCompareA(fpgaInfo.Identifier, pBitstream->Identifier) != 0) {
    status = ADMXRC3_FPGA_MISMATCH;
    goto done;
  }

  ioctl.in.targetIndex = targetIndex;
  ioctl.in.flags = flags;
  ioctl.in.pBuffer = pBitstream->Data;
  ioctl.in.length = pBitstream->Length;
  status = admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_CONFIGURE, &ioctl, sizeof(ioctl.in), 0);

done:
  if (NULL != pBitstream) {
    ADMXRC3_UnloadBitstreamA(pBitstream);
  }
  return status;
}

#if PORTABLE_HAS_WCHAR_SUPPORT
ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_ConfigureFromFileW(
  ADMXRC3_HANDLE hCard,
  unsigned int targetIndex,
  uint32_t flags,
  const wchar_t* pFilename)
{
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  ADMXRC3_BITSTREAMW* pBitstream = NULL;
  ADMXRC3_FPGA_INFOW fpgaInfo;
  IOCTLS_ADMXRC3_CONFIGURE ioctl;

  status = ADMXRC3_LoadBitstreamW(pFilename, &pBitstream);
  if (ADMXRC3_SUCCESS != status) {
    goto done;
  }

  status = ADMXRC3_GetFpgaInfoW(hCard, targetIndex, &fpgaInfo);
  if (ADMXRC3_SUCCESS != status) {
    goto done;
  }

  if (portableStringCompareW(fpgaInfo.Identifier, pBitstream->Identifier) != 0) {
    status = ADMXRC3_FPGA_MISMATCH;
    goto done;
  }

  ioctl.in.targetIndex = targetIndex;
  ioctl.in.flags = flags;
  ioctl.in.pBuffer = pBitstream->Data;
  ioctl.in.length = pBitstream->Length;
  status = admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_CONFIGURE, &ioctl, sizeof(ioctl.in), 0);

done:
  if (NULL != pBitstream) {
    ADMXRC3_UnloadBitstreamW(pBitstream);
  }
  return status;
}
#endif

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_EraseFlash(
  ADMXRC3_HANDLE hCard,
  unsigned int flashIndex,
  uint32_t flags,
  uint64_t offset,
  uint64_t length)
{
  IOCTLS_ADMXRC3_ERASEFLASH ioctl;

  ioctl.in.flashIndex = flashIndex;
  ioctl.in.flags = flags;
  ioctl.in.offset = offset;
  ioctl.in.length = length;
  return admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_ERASEFLASH, &ioctl, sizeof(ioctl.in), 0);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_FinishDMA(
  ADMXRC3_HANDLE hCard,
  ADMXRC3_TICKET* pTicket,
  bool_t bWait)
{
  return admxrc3FinishNonBlock(hCard, pTicket, bWait, NULL, 0);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_FinishNotificationWait(
  ADMXRC3_HANDLE hCard,
  ADMXRC3_TICKET* pTicket,
  bool_t bWait)
{
  return admxrc3FinishNonBlock(hCard, pTicket, bWait, NULL, 0);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_GetBankInfo(
  ADMXRC3_HANDLE hCard,
  unsigned int bankIndex,
  ADMXRC3_BANK_INFO* pBankInfo)
{
  IOCTLS_ADMXRC3_GETBANKINFO ioctl;
  ADMXRC3_STATUS status;

  if (NULL == pBankInfo) {
    return ADMXRC3_NULL_POINTER;
  }

  ioctl.in.bankIndex = bankIndex;
  status = admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_GETBANKINFO, &ioctl, sizeof(ioctl.in), sizeof(ioctl.out));
  if (ADMXRC3_SUCCESS == status) {
    memset(pBankInfo, 0, sizeof(*pBankInfo));
    pBankInfo->MaximumFrequency = ioctl.out.maximumFrequency;
    pBankInfo->MinimumFrequency = ioctl.out.minimumFrequency;
    pBankInfo->PhysicalSize = ioctl.out.physicalSize;
    pBankInfo->PhysicalDataWidth = ioctl.out.physicalDataWidth;
    pBankInfo->PhysicalECCWidth = ioctl.out.physicalEccWidth;
    pBankInfo->PhysicalWidth = (uint16_t)(pBankInfo->PhysicalDataWidth + pBankInfo->PhysicalECCWidth);
    pBankInfo->TypeMask = ioctl.out.typeMask;
    pBankInfo->ConnectivityMask = ioctl.out.connectivityMask;
    pBankInfo->Present = ioctl.out.bPresent ? TRUE : FALSE;
  }
  return status;
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_GetCardInfo(
  ADMXRC3_HANDLE hCard,
  ADMXRC3_CARD_INFO* pCardInfo)
{
  IOCTLS_ADMXRC3_GETCARDINFO ioctl;
  ADMXRC3_STATUS status;

  if (NULL == pCardInfo) {
    return ADMXRC3_NULL_POINTER;
  }

  status = admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_GETCARDINFO, &ioctl, 0, sizeof(ioctl.out));
  if (ADMXRC3_SUCCESS == status) {
    memset(pCardInfo, 0, sizeof(*pCardInfo));
    pCardInfo->Model = (ADMXRC3_MODEL_TYPE)ioctl.out.model;
    pCardInfo->SerialNumber = (uint32_t)ioctl.out.serialNumber;
    pCardInfo->NumClockGen = (unsigned int)ioctl.out.numClockGen;
    pCardInfo->NumDmaChannel = (unsigned int)ioctl.out.numDmaChannel;
    pCardInfo->NumFlashBank = (unsigned int)ioctl.out.numFlashBank;
    pCardInfo->NumMemoryBank = (unsigned int)ioctl.out.numMemoryBank;
    pCardInfo->NumTargetFpga = (unsigned int)ioctl.out.numTargetFpga;
    pCardInfo->NumWindow = (unsigned int)ioctl.out.numWindow;
    pCardInfo->MemoryBankPresent = (uint32_t)ioctl.out.memoryBankPresent;
  }
  return status;
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_GetCardInfoEx(
  ADMXRC3_HANDLE hCard,
  ADMXRC3_CARD_INFOEX* pCardInfoEx)
{
  IOCTLS_ADMXRC3_GETCARDINFOEX ioctl;
  ADMXRC3_STATUS status;

  if (NULL == pCardInfoEx) {
    return ADMXRC3_NULL_POINTER;
  }

  status = admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_GETCARDINFOEX, &ioctl, 0, sizeof(ioctl.out));
  if (ADMXRC3_SUCCESS == status) {
    memset(pCardInfoEx, 0, sizeof(*pCardInfoEx));
    pCardInfoEx->Model = (ADMXRC3_MODEL_TYPE)ioctl.out.model;
    pCardInfoEx->SerialNumber = (uint32_t)ioctl.out.serialNumber;
    pCardInfoEx->NumClockGen = (unsigned int)ioctl.out.numClockGen;
    pCardInfoEx->NumDmaChannel = (unsigned int)ioctl.out.numDmaChannel;
    pCardInfoEx->NumFlashBank = (unsigned int)ioctl.out.numFlashBank;
    pCardInfoEx->NumMemoryBank = (unsigned int)ioctl.out.numMemoryBank;
    pCardInfoEx->NumTargetFpga = (unsigned int)ioctl.out.numTargetFpga;
    pCardInfoEx->NumWindow = (unsigned int)ioctl.out.numWindow;
    pCardInfoEx->MemoryBankPresent = (uint32_t)ioctl.out.memoryBankPresent;
    pCardInfoEx->NumSensor = (uint32_t)ioctl.out.numSensor;
    pCardInfoEx->NumModuleSite = (uint32_t)ioctl.out.numModuleSite;
  }
  return status;
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_GetClockFrequency(
  ADMXRC3_HANDLE hCard,
  unsigned int clockIndex,
  double* pActualFrequency)
{
  IOCTLS_ADMXRC3_GETCLOCKFREQUENCY ioctl;
  ADMXRC3_STATUS status;

  ioctl.in.clockIndex = (uint32_t)clockIndex;
  status = admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_GETCLOCKFREQUENCY, &ioctl, sizeof(ioctl.in), sizeof(ioctl.out));
  if (ADMXRC3_SUCCESS == status) {
    if (NULL == pActualFrequency) {
      return ADMXRC3_NULL_POINTER;
    } else {
      /* Casting to int64_t then double avoids missing '__floadundidf' symbol when linking apps against shared library with some GNU toolchains */
       *pActualFrequency = (double)(int64_t)ioctl.out.frequency; /* This cast is safe, unless your hardware is generating X-Rays */
    }
  }
  return status;
}

#if 0
ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_GetCooperativeLevel(
  ADMXRC3_HANDLE hCard,
  uint32_t* pCooperativeLevel)
{
  IOCTLS_ADMXRC3_GETCOOPLEVEL ioctl;
  ADMXRC3_STATUS status;

  if (NULL == pCooperativeLevel) {
    return ADMXRC3_NULL_POINTER;
  }

  status = admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_GETCOOPLEVEL, &ioctl, 0, sizeof(ioctl.out));
  if (ADMXRC3_SUCCESS == status) {
    *pCooperativeLevel = ioctl.out.cooperativeLevel;
  }
  return status;
}
#endif

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_GetFlashBlockInfo(
  ADMXRC3_HANDLE hCard,
  unsigned int flashIndex,
  uint64_t address,
  ADMXRC3_FLASHBLOCK_INFO* pFlashBlockInfo)
{
  IOCTLS_ADMXRC3_GETFLASHBLOCKINFO ioctl;
  ADMXRC3_STATUS status;

  if (NULL == pFlashBlockInfo) {
    return ADMXRC3_NULL_POINTER;
  }

  ioctl.in.flashIndex = flashIndex;
  ioctl.in.address = address;
  status = admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_GETFLASHBLOCKINFO, &ioctl, sizeof(ioctl.in), sizeof(ioctl.out));
  if (ADMXRC3_SUCCESS == status) {
    pFlashBlockInfo->Address = ioctl.out.address;
    pFlashBlockInfo->Length = ioctl.out.length;
    pFlashBlockInfo->Flags = ioctl.out.flags;
  }
  return status;
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_GetFlashInfoA(
  ADMXRC3_HANDLE hCard,
  unsigned int flashIndex,
  ADMXRC3_FLASH_INFOA* pFlashInfo)
{
  IOCTLS_ADMXRC3_GETFLASHINFO ioctl;
  ADMXRC3_STATUS status;
  const char* pIdentifier = NULL;

  if (NULL == pFlashInfo) {
    return ADMXRC3_NULL_POINTER;
  }

  ioctl.in.flashIndex = flashIndex;
  status = admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_GETFLASHINFO, &ioctl, sizeof(ioctl.in), sizeof(ioctl.out));
  if (ADMXRC3_SUCCESS == status) {
    switch (ioctl.out.vendorId) {
    case 0x89U:
      /* Intel or Micron */
      switch (ioctl.out.deviceId) {
      case 0x99U:
        pIdentifier = "Micron 28F008B";
        break;

    	case 0x16U:
        pIdentifier = "Intel 28F320J3A";
        break;

    	case 0x17U:
        pIdentifier = "Intel 28F640J3A";
        break;

  	  case 0x18U:
        pIdentifier = "Intel 28F128J3A";
        break;

    	case 0x8803:
        pIdentifier = "Intel 28F256K3";
        break;

    	case 0x88B0:
        pIdentifier = "Micron MT28GU01GAAA1E";
        break;

      case 0x8919: /* Intel 28F256P30 or P30 stacked die flash, top boot */
        pIdentifier = "Intel 28F256P30";
        break;

      case 0x891C: /* Intel 28F256P30 or P30 stacked die flash, bottom boot */
        pIdentifier = "Intel 28F256P30";
        break;

      case 0x8999: /* Numonyx Axcell P30 512GBit flash, symmetric blocks */
      case 0x899A: /* Numonyx Axcell P30 1GBit flash, symmetric blocks */
        pIdentifier = "Numonyx Axcell P30 (Symm bl)";
        break;

      case 0x8960: /* Numonyx Axcell P30 512GBit flash, top boot */
      case 0x8962: /* Numonyx Axcell P30 1GBit flash, Top boot */
        pIdentifier = "Numonyx Axcell P30 (Top boot)";
        break;

      case 0x8961: /* Numonyx Axcell P30 512GBit flash, bottom boot */
      case 0x8963: /* Numonyx Axcell P30 1GBit flash, bottom boot */

        pIdentifier = "Numonyx Axcell P30 (Bot boot)";
        break;
      }
      break;

    case 0xbfU:
      /* SST */
      switch (ioctl.out.deviceId) {
      case 0xd9U:
        pIdentifier = "SST 39VF016Q";
        break;
      }
      break;
    }
    if (NULL == pIdentifier) {
      portableFormatStringA2(pFlashInfo->Identifier, ARRAY_LENGTH(pFlashInfo->Identifier), "Vendor 0x%04lX Device 0x%04lX",
        (unsigned long)ioctl.out.vendorId, (unsigned long)ioctl.out.deviceId);
    } else {
      portableStringCopy(pFlashInfo->Identifier, ARRAY_LENGTH(pFlashInfo->Identifier), pIdentifier);
    }
    pFlashInfo->Size = ioctl.out.size;
    pFlashInfo->UseableStart = ioctl.out.useableStart;
    pFlashInfo->UseableLength = ioctl.out.useableLength;
  }
  return status;
}

#if PORTABLE_HAS_WCHAR_SUPPORT
ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_GetFlashInfoW(
  ADMXRC3_HANDLE hCard,
  unsigned int flashIndex,
  ADMXRC3_FLASH_INFOW* pFlashInfo)
{
  ADMXRC3_FLASH_INFOA flashInfoA;
  ADMXRC3_STATUS status;

  if (NULL == pFlashInfo) {
    return ADMXRC3_NULL_POINTER;
  }

  status = ADMXRC3_GetFlashInfoA(hCard, flashIndex, &flashInfoA);
  if (ADMXRC3_SUCCESS == status) {
    portableUTF8ToWideChar(pFlashInfo->Identifier, ARRAY_LENGTH(pFlashInfo->Identifier), flashInfoA.Identifier);
    pFlashInfo->Size = flashInfoA.Size;
    pFlashInfo->UseableStart = flashInfoA.UseableStart;
    pFlashInfo->UseableLength = flashInfoA.UseableLength;
  }
  return status;
}
#endif

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_GetFpgaInfoA(
  ADMXRC3_HANDLE hCard,
  unsigned int targetIndex,
  ADMXRC3_FPGA_INFOA* pFpgaInfo)
{
  IOCTLS_ADMXRC3_GETFPGAINFO ioctl;
  ADMXRC3_STATUS status;
  char deviceName[32];
  char packageName[32];

  if (NULL == pFpgaInfo) {
    return ADMXRC3_NULL_POINTER;
  }

  ioctl.in.targetIndex = targetIndex;
  status = admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_GETFPGAINFO, &ioctl, sizeof(ioctl.in), sizeof(ioctl.out));
  if (ADMXRC3_SUCCESS == status) {
    memset(pFpgaInfo, 0, sizeof(*pFpgaInfo));
    if (ioctl.out.deviceCode >= ARRAY_LENGTH(fpgaDeviceNamesA)) {
      portableFormatStringA1(deviceName, ARRAY_LENGTH(deviceName), "unknown%lx", (unsigned long)ioctl.out.deviceCode);
    } else {
      portableStringCopy(deviceName, ARRAY_LENGTH(deviceName), fpgaDeviceNamesA[ioctl.out.deviceCode]);
    }
    if (ioctl.out.familyCode == ADMXRC3_FAMILY_7SERIES) {
      /* 7 series has package type FBG, FFG etc. instead of FF, FB etc. where G means lead-free */
      portableFormatStringA3(packageName, ARRAY_LENGTH(packageName), "%c%cg%lu",
        tolower(ADMXRC3_PACKAGE_GETTYPE0(ioctl.out.packageCode)),
        tolower(ADMXRC3_PACKAGE_GETTYPE1(ioctl.out.packageCode)),
        (unsigned long)ADMXRC3_PACKAGE_GETPINS(ioctl.out.packageCode));
    } else if (ioctl.out.familyCode == ADMXRC3_FAMILY_ULTRASCALE) {
      /*
      ** Ultrascale packages are defined in Xilinx UG575.
      ** We have only two bytes (ASCII chars) to map to a a 4-letter package, so:
      ** - 1st letter of package is always F (flip-chip).
      ** - 2nd & 3rd letters are obtained from ADMXRC3_PACKAGE_GETTYPE0:
      **     byte 0   |  2nd    3rd
      **   - B        |  B      V
      **   - b        |  B      G
      **   - F        |  F      V
      **   - f        |  F      G
      **   - L        |  L      V
      **   - l        |  L      G
      ** - 4th letter is obtained directly from ADMXRC3_PACKAGE_GETTYPE1.
      **
      ** NOTE: the string returned in the packageName field must be converted to lower case.
      */
      char package[5];
      uint8_t byte0, byte1;

      byte0 = ADMXRC3_PACKAGE_GETTYPE0(ioctl.out.packageCode);
      byte1 = ADMXRC3_PACKAGE_GETTYPE1(ioctl.out.packageCode);
      package[0] = 'f';
      package[1] = (char)tolower((char)byte0);
      if (islower(byte0)) {
        package[2] = 'g';
      } else {
        package[2] = 'v';
      }
      package[3] = (char)tolower((char)byte1);
      package[4] = '\0';
      portableFormatStringA2(packageName, ARRAY_LENGTH(packageName), "%s%lu",
        package, (unsigned long)ADMXRC3_PACKAGE_GETPINS(ioctl.out.packageCode));
    } else {
      portableFormatStringA3(packageName, ARRAY_LENGTH(packageName), "%c%c%lu",
        tolower(ADMXRC3_PACKAGE_GETTYPE0(ioctl.out.packageCode)),
        tolower(ADMXRC3_PACKAGE_GETTYPE1(ioctl.out.packageCode)),
        (unsigned long)ADMXRC3_PACKAGE_GETPINS(ioctl.out.packageCode));
    }
    portableStringCopy(pFpgaInfo->Identifier, ARRAY_LENGTH(pFpgaInfo->Identifier), deviceName);
    portableStringCat(pFpgaInfo->Identifier, ARRAY_LENGTH(pFpgaInfo->Identifier), packageName);
    pFpgaInfo->DeviceCode = (ADMXRC3_FPGA_TYPE)ioctl.out.deviceCode;
    pFpgaInfo->FamilyCode = (ADMXRC3_FAMILY_TYPE)ioctl.out.familyCode;
    pFpgaInfo->SubfamilyCode = (ADMXRC3_SUBFAMILY_TYPE)ioctl.out.subfamilyCode;
    pFpgaInfo->PackageCode = ioctl.out.packageCode;
    portableStringUpper(deviceName, ARRAY_LENGTH(deviceName));
    portableStringCopy(pFpgaInfo->Device, ARRAY_LENGTH(pFpgaInfo->Device), deviceName);
    portableStringUpper(packageName, ARRAY_LENGTH(packageName));
    portableStringCopy(pFpgaInfo->Package, ARRAY_LENGTH(pFpgaInfo->Package), packageName);
    pFpgaInfo->Flags = ioctl.out.flags;
    portableStringCopy(pFpgaInfo->SpeedGrade, ARRAY_LENGTH(pFpgaInfo->SpeedGrade), ioctl.out.speedGrade);
    portableStringCopy(pFpgaInfo->Stepping, ARRAY_LENGTH(pFpgaInfo->Stepping), ioctl.out.stepping);
    pFpgaInfo->Present = ioctl.out.bPresent;
  }
  return status;
}

#if PORTABLE_HAS_WCHAR_SUPPORT
ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_GetFpgaInfoW(
  ADMXRC3_HANDLE hCard,
  unsigned int targetIndex,
  ADMXRC3_FPGA_INFOW* pFpgaInfo)
{
  ADMXRC3_FPGA_INFOA fpgaInfoA;
  ADMXRC3_STATUS status;

  if (NULL == pFpgaInfo) {
    return ADMXRC3_NULL_POINTER;
  }

  status = ADMXRC3_GetFpgaInfoA(hCard, targetIndex, &fpgaInfoA);
  if (ADMXRC3_SUCCESS == status) {
    portableUTF8ToWideChar(pFpgaInfo->Identifier, ARRAY_LENGTH(pFpgaInfo->Identifier), fpgaInfoA.Identifier);
    pFpgaInfo->DeviceCode = fpgaInfoA.DeviceCode;
    pFpgaInfo->FamilyCode = fpgaInfoA.FamilyCode;
    pFpgaInfo->SubfamilyCode = fpgaInfoA.SubfamilyCode;
    pFpgaInfo->PackageCode = fpgaInfoA.PackageCode;
    portableUTF8ToWideChar(pFpgaInfo->Device, ARRAY_LENGTH(pFpgaInfo->Device), fpgaInfoA.Device);
    portableUTF8ToWideChar(pFpgaInfo->Package, ARRAY_LENGTH(pFpgaInfo->Package), fpgaInfoA.Package);
    pFpgaInfo->Flags = fpgaInfoA.Flags;
    portableUTF8ToWideChar(pFpgaInfo->SpeedGrade, ARRAY_LENGTH(pFpgaInfo->SpeedGrade), fpgaInfoA.SpeedGrade);
    portableUTF8ToWideChar(pFpgaInfo->Stepping, ARRAY_LENGTH(pFpgaInfo->Stepping), fpgaInfoA.Stepping);
    pFpgaInfo->Present = fpgaInfoA.Present;
  }
  return status;
}
#endif

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_GetModuleInfoA(
  ADMXRC3_HANDLE hCard,
  unsigned int moduleIndex,
  ADMXRC3_MODULE_INFOA* pModuleInfo)
{
  IOCTLS_ADMXRC3_GETMODULEINFO ioctl;
  IOCTLSUBS_ADMXRC3_GETMODULEINFO subs;
  ADMXRC3_STATUS status;

  if (NULL == pModuleInfo) {
    return ADMXRC3_NULL_POINTER;
  }

  ioctl.in.moduleIndex = moduleIndex;
  ioctl.in.pBuffer = &subs;
  status = admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_GETMODULEINFO, &ioctl, sizeof(ioctl.in), 0);
  if (ADMXRC3_SUCCESS == status) {
    memset(pModuleInfo, 0, sizeof(*pModuleInfo));
    portableStringCopy(pModuleInfo->Product, ARRAY_LENGTH(pModuleInfo->Product), subs.product);
    portableStringCopy(pModuleInfo->PartNumber, ARRAY_LENGTH(pModuleInfo->PartNumber), subs.partNumber);
    portableStringCopy(pModuleInfo->Manufacturer, ARRAY_LENGTH(pModuleInfo->Manufacturer), subs.manufacturer);
    portableStringCopy(pModuleInfo->SerialNumber, ARRAY_LENGTH(pModuleInfo->SerialNumber), subs.serialNumber);
    pModuleInfo->IoVoltage = TO_DOUBLE32(subs.ioVoltage.intPart, subs.ioVoltage.fracPart);
    pModuleInfo->ManufactureDate = subs.manufactureDate;
    pModuleInfo->Flags = subs.flags;
    pModuleInfo->Present = subs.bPresent ? TRUE : FALSE;
  }
  return status;
}

#if PORTABLE_HAS_WCHAR_SUPPORT
ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_GetModuleInfoW(
  ADMXRC3_HANDLE hCard,
  unsigned int moduleIndex,
  ADMXRC3_MODULE_INFOW* pModuleInfo)
{
  ADMXRC3_MODULE_INFOA moduleInfoA;
  ADMXRC3_STATUS status;

  if (NULL == pModuleInfo) {
    return ADMXRC3_NULL_POINTER;
  }

  status = ADMXRC3_GetModuleInfoA(hCard, moduleIndex, &moduleInfoA);
  if (ADMXRC3_SUCCESS == status) {
    portableUTF8ToWideChar(pModuleInfo->Product, ARRAY_LENGTH(pModuleInfo->Product), moduleInfoA.Product);
    portableUTF8ToWideChar(pModuleInfo->PartNumber, ARRAY_LENGTH(pModuleInfo->PartNumber), moduleInfoA.PartNumber);
    portableUTF8ToWideChar(pModuleInfo->Manufacturer, ARRAY_LENGTH(pModuleInfo->Manufacturer), moduleInfoA.Manufacturer);
    portableUTF8ToWideChar(pModuleInfo->SerialNumber, ARRAY_LENGTH(pModuleInfo->SerialNumber), moduleInfoA.SerialNumber);
    pModuleInfo->IoVoltage = moduleInfoA.IoVoltage;
    pModuleInfo->ManufactureDate = moduleInfoA.ManufactureDate;
    pModuleInfo->Flags = moduleInfoA.Flags;
    pModuleInfo->Present = moduleInfoA.Present;
  }
  return status;
}
#endif

ADMXRC3_EXPORT const char*
ADMXRC3_CALLING_CONVENTION
ADMXRC3_GetStatusStringA(
  ADMXRC3_STATUS code,
  bool_t bShort)
{
  static char buffer[128];

  if (bShort) {
    switch (code) {
    case ADMXRC3_SUCCESS:
      return "ADMXRC3_SUCCESS";

    case ADMXRC3_INTERNAL_ERROR:
      return "ADMXRC3_INTERNAL_ERROR";

    case ADMXRC3_UNEXPECTED_ERROR:
      return "ADMXRC3_UNEXPECTED_ERROR";

    case ADMXRC3_BAD_DRIVER:
      return "ADMXRC3_BAD_DRIVER";

    case ADMXRC3_NO_MEMORY:
      return "ADMXRC3_NO_MEMORY";

    case ADMXRC3_ACCESS_DENIED:
      return "ADMXRC3_ACCESS_DENIED";

    case ADMXRC3_DEVICE_NOT_FOUND:
      return "ADMXRC3_DEVICE_NOT_FOUND";

    case ADMXRC3_FILE_NOT_FOUND:
      return "ADMXRC3_FILE_NOT_FOUND";

    case ADMXRC3_HARDWARE_ERROR:
      return "ADMXRC3_HARDWARE_ERROR";

    case ADMXRC3_FPGA_MISMATCH:
      return "ADMXRC3_FPGA_MISMATCH";

    case ADMXRC3_INVALID_BITSTREAM:
      return "ADMXRC3_INVALID_BITSTREAM";

    case ADMXRC3_INVALID_BUFFER:
      return "ADMXRC3_INVALID_BUFFER";

    case ADMXRC3_INVALID_FLAG:
      return "ADMXRC3_INVALID_FLAG";

    case ADMXRC3_INVALID_FREQUENCY:
      return "ADMXRC3_INVALID_FREQUENCY";

    case ADMXRC3_INVALID_HANDLE:
      return "ADMXRC3_INVALID_HANDLE";

    case ADMXRC3_INVALID_INDEX:
      return "ADMXRC3_INVALID_INDEX";

    case ADMXRC3_INVALID_REGION:
      return "ADMXRC3_INVALID_REGION";

    case ADMXRC3_NULL_POINTER:
      return "ADMXRC3_NULL_POINTER";

    case ADMXRC3_BUSY:
      return "ADMXRC3_BUSY";

    case ADMXRC3_CANCEL_FAILED:
      return "ADMXRC3_CANCEL_FAILED";

    case ADMXRC3_CANCELLED:
      return "ADMXRC3_CANCELLED";

    case ADMXRC3_RESOURCE_LIMIT:
      return "ADMXRC3_RESOURCE_LIMIT";

    case ADMXRC3_REGION_TOO_LARGE:
      return "ADMXRC3_REGION_TOO_LARGE";

    case ADMXRC3_MUST_WAIT:
      return "ADMXRC3_MUST_WAIT";

    case ADMXRC3_NOT_OWNER:
      return "ADMXRC3_NOT_OWNER";

    case ADMXRC3_PENDING:
      return "ADMXRC3_PENDING";

    case ADMXRC3_NONBLOCK_IDLE:
      return "ADMXRC3_NONBLOCK_IDLE";

    case ADMXRC3_INVALID_BUFFER_HANDLE:
      return "ADMXRC3_INVALID_BUFFER_HANDLE";

    case ADMXRC3_INVALID_LOCAL_REGION:
      return "ADMXRC3_INVALID_LOCAL_REGION";

    case ADMXRC3_DATATYPE_MISMATCH:
      return "ADMXRC3_DATATYPE_MISMATCH";

    case ADMXRC3_NOT_SUPPORTED:
      return "ADMXRC3_NOT_SUPPORTED";

    default:
      break;
    }
  } else {
    switch (code) {
    case ADMXRC3_SUCCESS:
      return "No error";

    case ADMXRC3_INTERNAL_ERROR:
      return "An error in the API logic was detected";

    case ADMXRC3_UNEXPECTED_ERROR:
      return "An unexpected error occurred";

    case ADMXRC3_BAD_DRIVER:
      return "The driver may be incorrectly installed";

    case ADMXRC3_NO_MEMORY:
      return "Could not allocate memory required to complete operation";

    case ADMXRC3_ACCESS_DENIED:
      return "Caller does not have permission to perform the operation";

    case ADMXRC3_DEVICE_NOT_FOUND:
      return "Failed to open the device with specified CardID";

    case ADMXRC3_FILE_NOT_FOUND:
      return "Failed to open the specified bitstream file";

    case ADMXRC3_HARDWARE_ERROR:
      return "A hardware error occurred";

    case ADMXRC3_FPGA_MISMATCH:
      return "The FPGA on the card did not match that of the bitstream file";

    case ADMXRC3_INVALID_BITSTREAM:
      return "The bitstream file was corrupt";

    case ADMXRC3_INVALID_BUFFER:
      return "The supplied buffer was invalid";

    case ADMXRC3_INVALID_FLAG:
      return "A flag was invalid or not recognized";

    case ADMXRC3_INVALID_FREQUENCY:
      return "The frequency was invalid";

    case ADMXRC3_INVALID_HANDLE:
      return "The device handle was invalid";

    case ADMXRC3_INVALID_INDEX:
      return "The index parameter was invalid";

    case ADMXRC3_INVALID_REGION:
      return "The offset and/or length parameter(s) were invalid";

    case ADMXRC3_NULL_POINTER:
      return "A NULL pointer was passed where NULL was illegal";

    case ADMXRC3_BUSY:
      return "There is already an operation in progress";

    case ADMXRC3_CANCEL_FAILED:
      return "There is nothing to cancel";

    case ADMXRC3_CANCELLED:
      return "The operation was cancelled";

    case ADMXRC3_RESOURCE_LIMIT:
      return "A resource limit was reached";

    case ADMXRC3_REGION_TOO_LARGE:
      return "The region is too large for a single operation";

    case ADMXRC3_MUST_WAIT:
      return "The operation could not be started immediately";

    case ADMXRC3_NOT_OWNER:
      return "The buffer handle belongs to a different device handle";

    case ADMXRC3_PENDING:
      return "The non-blocking operation is now in progress";

    case ADMXRC3_NONBLOCK_IDLE:
      return "There is no non-blocking operation to finish";

    case ADMXRC3_INVALID_BUFFER_HANDLE:
      return "The buffer handle is invalid";

    case ADMXRC3_INVALID_LOCAL_REGION:
      return "The local address region is invalid";

    case ADMXRC3_DATATYPE_MISMATCH:
      return "The datatype did not match that of the sensor";

    case ADMXRC3_NOT_SUPPORTED:
      return "The hardware does not support the requested operation";

    default:
      break;
    }
  }
  portableFormatStringA2(buffer, ARRAY_LENGTH(buffer), "Unknown status code %lu(0x%lx)", (unsigned long)code, (unsigned long)code);
  return buffer;
}

#if PORTABLE_HAS_WCHAR_SUPPORT
ADMXRC3_EXPORT const wchar_t*
ADMXRC3_CALLING_CONVENTION
ADMXRC3_GetStatusStringW(
  ADMXRC3_STATUS code,
  bool_t bShort)
{
  static char bufferA[128];
  static wchar_t bufferW[128];

  if (bShort) {
    switch (code) {
    case ADMXRC3_SUCCESS:
      return L"ADMXRC3_SUCCESS";

    case ADMXRC3_INTERNAL_ERROR:
      return L"ADMXRC3_INTERNAL_ERROR";

    case ADMXRC3_UNEXPECTED_ERROR:
      return L"ADMXRC3_UNEXPECTED_ERROR";

    case ADMXRC3_BAD_DRIVER:
      return L"ADMXRC3_BAD_DRIVER";

    case ADMXRC3_NO_MEMORY:
      return L"ADMXRC3_NO_MEMORY";

    case ADMXRC3_ACCESS_DENIED:
      return L"ADMXRC3_ACCESS_DENIED";

    case ADMXRC3_DEVICE_NOT_FOUND:
      return L"ADMXRC3_DEVICE_NOT_FOUND";

    case ADMXRC3_FILE_NOT_FOUND:
      return L"ADMXRC3_FILE_NOT_FOUND";

    case ADMXRC3_HARDWARE_ERROR:
      return L"ADMXRC3_HARDWARE_ERROR";

    case ADMXRC3_FPGA_MISMATCH:
      return L"ADMXRC3_FPGA_MISMATCH";

    case ADMXRC3_INVALID_BITSTREAM:
      return L"ADMXRC3_INVALID_BITSTREAM";

    case ADMXRC3_INVALID_BUFFER:
      return L"ADMXRC3_INVALID_BUFFER";

    case ADMXRC3_INVALID_FLAG:
      return L"ADMXRC3_INVALID_FLAG";

    case ADMXRC3_INVALID_FREQUENCY:
      return L"ADMXRC3_INVALID_FREQUENCY";

    case ADMXRC3_INVALID_HANDLE:
      return L"ADMXRC3_INVALID_HANDLE";

    case ADMXRC3_INVALID_INDEX:
      return L"ADMXRC3_INVALID_INDEX";

    case ADMXRC3_INVALID_REGION:
      return L"ADMXRC3_INVALID_REGION";

    case ADMXRC3_NULL_POINTER:
      return L"ADMXRC3_NULL_POINTER";

    case ADMXRC3_BUSY:
      return L"ADMXRC3_BUSY";

    case ADMXRC3_CANCEL_FAILED:
      return L"ADMXRC3_CANCEL_FAILED";

    case ADMXRC3_CANCELLED:
      return L"ADMXRC3_CANCELLED";

    case ADMXRC3_RESOURCE_LIMIT:
      return L"ADMXRC3_RESOURCE_LIMIT";

    case ADMXRC3_REGION_TOO_LARGE:
      return L"ADMXRC3_REGION_TOO_LARGE";

    case ADMXRC3_MUST_WAIT:
      return L"ADMXRC3_MUST_WAIT";

    case ADMXRC3_NOT_OWNER:
      return L"ADMXRC3_NOT_OWNER";

    case ADMXRC3_PENDING:
      return L"ADMXRC3_PENDING";

    case ADMXRC3_NONBLOCK_IDLE:
      return L"ADMXRC3_NONBLOCK_IDLE";

    case ADMXRC3_INVALID_BUFFER_HANDLE:
      return L"ADMXRC3_INVALID_BUFFER_HANDLE";

    case ADMXRC3_INVALID_LOCAL_REGION:
      return L"ADMXRC3_INVALID_LOCAL_REGION";

    case ADMXRC3_DATATYPE_MISMATCH:
      return L"ADMXRC3_DATATYPE_MISMATCH";

    case ADMXRC3_NOT_SUPPORTED:
      return L"ADMXRC3_NOT_SUPPORTED";

    default:
      break;
    }
  } else {
    switch (code) {
    case ADMXRC3_SUCCESS:
      return L"No error";

    case ADMXRC3_INTERNAL_ERROR:
      return L"An error in the API logic was detected";

    case ADMXRC3_UNEXPECTED_ERROR:
      return L"An unexpected error occurred";

    case ADMXRC3_BAD_DRIVER:
      return L"The driver may be incorrectly installed";

    case ADMXRC3_NO_MEMORY:
      return L"Could not allocate memory required to complete operation";

    case ADMXRC3_ACCESS_DENIED:
      return L"Caller does not have permission to perform the operation";

    case ADMXRC3_DEVICE_NOT_FOUND:
      return L"Failed to open the device with specified CardID";

    case ADMXRC3_FILE_NOT_FOUND:
      return L"Failed to open the specified bitstream file";

    case ADMXRC3_HARDWARE_ERROR:
      return L"A hardware error occurred";

    case ADMXRC3_FPGA_MISMATCH:
      return L"The FPGA on the card did not match that of the bitstream file";

    case ADMXRC3_INVALID_BITSTREAM:
      return L"The bitstream file was corrupt";

    case ADMXRC3_INVALID_BUFFER:
      return L"The supplied buffer was invalid";

    case ADMXRC3_INVALID_FLAG:
      return L"A flag was invalid or not recognized";

    case ADMXRC3_INVALID_FREQUENCY:
      return L"The frequency was invalid";

    case ADMXRC3_INVALID_HANDLE:
      return L"The device handle was invalid";

    case ADMXRC3_INVALID_INDEX:
      return L"The index parameter was invalid";

    case ADMXRC3_INVALID_REGION:
      return L"The offset and/or length parameter(s) were invalid";

    case ADMXRC3_NULL_POINTER:
      return L"A NULL pointer was passed where NULL was illegal";

    case ADMXRC3_BUSY:
      return L"There is already an operation in progress";

    case ADMXRC3_CANCEL_FAILED:
      return L"There is nothing to cancel";

    case ADMXRC3_CANCELLED:
      return L"The operation was cancelled";

    case ADMXRC3_RESOURCE_LIMIT:
      return L"A resource limit was reached";

    case ADMXRC3_REGION_TOO_LARGE:
      return L"The region is too large for a single operation";

    case ADMXRC3_MUST_WAIT:
      return L"The operation could not be started immediately";

    case ADMXRC3_NOT_OWNER:
      return L"The buffer handle belongs to a different device handle";

    case ADMXRC3_PENDING:
      return L"The non-blocking operation is now in progress";

    case ADMXRC3_NONBLOCK_IDLE:
      return L"There is no non-blocking operation to finish";

    case ADMXRC3_INVALID_BUFFER_HANDLE:
      return L"The buffer handle is invalid";

    case ADMXRC3_INVALID_LOCAL_REGION:
      return L"The local address region is invalid";

    case ADMXRC3_DATATYPE_MISMATCH:
      return L"The datatype did not match that of the sensor";

    case ADMXRC3_NOT_SUPPORTED:
      return L"The hardware does not support the requested operation";

    default:
      break;
    }
  }

  portableFormatStringA2(bufferA, ARRAY_LENGTH(bufferA), "Unknown status code %lu(0x%lx)", (unsigned long)code, (unsigned long)code);
  portableUTF8ToWideChar(bufferW, ARRAY_LENGTH(bufferW), bufferA);
  return bufferW;
}
#endif

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_GetVersionInfo(
  ADMXRC3_HANDLE hCard,
  ADMXRC3_VERSION_INFO* pVersionInfo)
{
  IOCTLS_ADMXRC3_GETDRIVERVERSION ioctl;
  ADMXRC3_STATUS status;

  if (NULL == pVersionInfo) {
    return ADMXRC3_NULL_POINTER;
  }

  status = admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_GETDRIVERVERSION, &ioctl, 0, sizeof(ioctl.out));
  if (ADMXRC3_SUCCESS == status) {
    pVersionInfo->Api.Major = ADMXRC3_VERSION_0;
    pVersionInfo->Api.Minor = ADMXRC3_VERSION_1;
    pVersionInfo->Api.Bugfix = ADMXRC3_VERSION_2;
    pVersionInfo->Driver.Major = ioctl.out.major;
    pVersionInfo->Driver.Minor = ioctl.out.minor;
    pVersionInfo->Driver.Bugfix = ioctl.out.bugfix;
  }
  return status;
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_GetWindowInfo(
  ADMXRC3_HANDLE hCard,
  unsigned int windowIndex,
  ADMXRC3_WINDOW_INFO* pWindowInfo)
{
  IOCTLS_ADMXRC3_GETWINDOWINFO ioctl;
  ADMXRC3_STATUS status;

  if (NULL == pWindowInfo) {
    return ADMXRC3_NULL_POINTER;
  }

  ioctl.in.windowIndex = windowIndex;
  status = admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_GETWINDOWINFO, &ioctl, sizeof(ioctl.in), sizeof(ioctl.out));
  if (ADMXRC3_SUCCESS == status) {
    memset(pWindowInfo, 0, sizeof(*pWindowInfo));
    pWindowInfo->BusBase = ioctl.out.busBase;
    pWindowInfo->BusSize = ioctl.out.busSize;
    pWindowInfo->LocalBase = ioctl.out.localBase;
    pWindowInfo->LocalSize = ioctl.out.localSize;
    pWindowInfo->VirtualSize = ioctl.out.virtualSize;
  }
  return status;
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_LoadBitstreamA(
  const char* pFilename,
  ADMXRC3_BITSTREAMA** ppBitstream)
{
  portable_file_t hFile;

  if (NULL == pFilename) {
    return ADMXRC3_NULL_POINTER;
  }
  if (NULL == ppBitstream) {
    return ADMXRC3_NULL_POINTER;
  }

  hFile = portableFileOpenA(pFilename, FALSE);
  if (hFile == PORTABLE_INVALID_FILE) {
    return ADMXRC3_FILE_NOT_FOUND;
  }

  return loadBitstreamA(hFile, ppBitstream);
}

#if PORTABLE_HAS_WCHAR_SUPPORT
ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_LoadBitstreamW(
  const wchar_t* pFilename,
  ADMXRC3_BITSTREAMW** ppBitstream)
{
  portable_file_t hFile;

  if (NULL == pFilename) {
    return ADMXRC3_NULL_POINTER;
  }
  if (NULL == ppBitstream) {
    return ADMXRC3_NULL_POINTER;
  }

  hFile = portableFileOpenW(pFilename, FALSE);
  if (hFile == PORTABLE_INVALID_FILE) {
    return ADMXRC3_FILE_NOT_FOUND;
  }

  return loadBitstreamW(hFile, ppBitstream);
}
#endif

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_Lock(
  ADMXRC3_HANDLE hCard,
  const void* pBuffer,
  size_t length,
  ADMXRC3_BUFFER_HANDLE* phBuffer)
{
  IOCTLS_ADMXRC3_LOCK ioctl;
  ADMXRC3_STATUS status;

  if (NULL == phBuffer) {
    return ADMXRC3_NULL_POINTER;
  }

  ioctl.in.pBuffer = pBuffer;
  ioctl.in.length = length;
  status = admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_LOCK, &ioctl, sizeof(ioctl.in), sizeof(ioctl.out));
  if (ADMXRC3_SUCCESS == status) {
    *phBuffer = ioctl.out.hBuffer;
  }
  return status;
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_MapWindow(
  ADMXRC3_HANDLE hCard,
  unsigned int windowIndex,
  uint64_t offset,
  uint64_t length,
  void** ppVirtualBase)
{
  ADMXRC3_CARD_INFO cardInfo;
  ADMXRC3_STATUS status;

  /* Sanity checks */
  if (0 == length || offset + length < offset) {
    return ADMXRC3_INVALID_REGION;
  }
  if (NULL == ppVirtualBase) {
    return ADMXRC3_NULL_POINTER;
  }
  /* Check for invalid window index */
  status = ADMXRC3_GetCardInfo(hCard, &cardInfo);
  if (ADMXRC3_SUCCESS != status) {
    return status;
  }
  if (windowIndex >= cardInfo.NumWindow) {
    return ADMXRC3_INVALID_INDEX;
  }

  /* This has to be done in an OS-specific way */
  return admxrc3MapRegion(hCard, windowIndex + ADMXRC3_REGION_TAG_WINDOW, offset, length, ppVirtualBase);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_Open(
  unsigned int index,
  ADMXRC3_HANDLE* phCard)
{
  if (NULL == phCard) {
    return ADMXRC3_NULL_POINTER;
  }
  /* This has to be done in an OS-specific way */
  return admxrc3OpenDevice(index, FALSE, phCard);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_OpenEx(
  unsigned int index,
  _ADMXRC3_BOOL bPassive,
  _ADMXRC3_UINT32 cooperativeLevel,
  ADMXRC3_HANDLE* phCard)
{
  ADMXRC3_STATUS status;
  ADMXRC3_HANDLE hCard = ADMXRC3_HANDLE_INVALID_VALUE;

  if (NULL == phCard) {
    return ADMXRC3_NULL_POINTER;
  }

  /* This has to be done in an OS-specific way */
  status = admxrc3OpenDevice(index, bPassive, &hCard);
  if (ADMXRC3_SUCCESS == status) {
    if (cooperativeLevel & ADMXRC3_COOP_EXCLUSIVE) {
      status = ADMXRC3_SetCooperativeLevel(hCard, cooperativeLevel);
      if (ADMXRC3_SUCCESS == status) {
        *phCard = hCard;
      } else {
        ADMXRC3_Close(hCard);
      }
    } else {
      *phCard = hCard;
    }
  }
  return status;
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_Read(
    ADMXRC3_HANDLE hCard,
    unsigned int windowIndex,
    uint32_t flags,
    size_t offset,
    size_t length,
    void* pBuffer)
{
  IOCTLS_ADMXRC3_READ ioctl;

  ioctl.in.windowIndex = windowIndex;
  ioctl.in.flags = flags;
  ioctl.in.offset = offset;
  ioctl.in.length = length;
  ioctl.in.pBuffer = pBuffer;
  return admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_READ, &ioctl, sizeof(ioctl.in), 0);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_ReadDMA(
  ADMXRC3_HANDLE hCard,
  unsigned int dmaChannel,
  uint32_t flags,
  void* pBuffer,
  size_t length,
  uint32_t localAddress)
{
  IOCTLS_ADMXRC3_READDMA ioctl;

  ioctl.in.channel = dmaChannel;
  ioctl.in.flags = flags;
  ioctl.in.pBuffer = pBuffer;
  ioctl.in.length = length;
  ioctl.in.local = localAddress;
  return admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_READDMA, &ioctl, sizeof(ioctl.in), 0);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_ReadDMAEx(
  ADMXRC3_HANDLE hCard,
  unsigned int dmaChannel,
  uint32_t flags,
  void* pBuffer,
  size_t length,
  uint64_t localAddress)
{
  IOCTLS_ADMXRC3_READDMA ioctl;

  ioctl.in.channel = dmaChannel;
  ioctl.in.flags = flags;
  ioctl.in.pBuffer = pBuffer;
  ioctl.in.length = length;
  ioctl.in.local = localAddress;
  return admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_READDMA, &ioctl, sizeof(ioctl.in), 0);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_ReadDMALocked(
  ADMXRC3_HANDLE hCard,
  unsigned int dmaChannel,
  uint32_t flags,
  ADMXRC3_BUFFER_HANDLE hBuffer,
  size_t offset,
  size_t length,
  uint32_t localAddress)
{
  IOCTLS_ADMXRC3_READDMALOCKED ioctl;

  ioctl.in.channel = dmaChannel;
  ioctl.in.flags = flags;
  ioctl.in.hBuffer = hBuffer;
  ioctl.in.offset = offset;
  ioctl.in.length = length;
  ioctl.in.local = localAddress;
  return admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_READDMALOCKED, &ioctl, sizeof(ioctl.in), 0);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_ReadDMALockedEx(
  ADMXRC3_HANDLE hCard,
  unsigned int dmaChannel,
  uint32_t flags,
  ADMXRC3_BUFFER_HANDLE hBuffer,
  size_t offset,
  size_t length,
  uint64_t localAddress)
{
  IOCTLS_ADMXRC3_READDMALOCKED ioctl;

  ioctl.in.channel = dmaChannel;
  ioctl.in.flags = flags;
  ioctl.in.hBuffer = hBuffer;
  ioctl.in.offset = offset;
  ioctl.in.length = length;
  ioctl.in.local = localAddress;
  return admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_READDMALOCKED, &ioctl, sizeof(ioctl.in), 0);
}

static ADMXRC3_STATUS
ADMXRC3_SetCooperativeLevel(
  ADMXRC3_HANDLE hCard,
  uint32_t cooperativeLevel)
{
  IOCTLS_ADMXRC3_SETCOOPLEVEL ioctl;

  ioctl.in.cooperativeLevel = cooperativeLevel;
  return admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_SETCOOPLEVEL, &ioctl, sizeof(ioctl.in), 0);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_StartReadDMA(
  ADMXRC3_HANDLE hCard,
  ADMXRC3_TICKET* pTicket,
  unsigned int dmaChannel,
  uint32_t flags,
  void* pBuffer,
  size_t length,
  uint32_t localAddress)
{
  IOCTLS_ADMXRC3_READDMA ioctl;

  ioctl.in.channel = dmaChannel;
  ioctl.in.flags = flags;
  ioctl.in.pBuffer = pBuffer;
  ioctl.in.length = length;
  ioctl.in.local = localAddress;
  return admxrc3IoctlNonBlock(hCard, pTicket, ADMXRC3_IOCTLCODE_READDMA, &ioctl, sizeof(ioctl.in), 0);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_StartReadDMAEx(
  ADMXRC3_HANDLE hCard,
  ADMXRC3_TICKET* pTicket,
  unsigned int dmaChannel,
  uint32_t flags,
  void* pBuffer,
  size_t length,
  uint64_t localAddress)
{
  IOCTLS_ADMXRC3_READDMA ioctl;

  ioctl.in.channel = dmaChannel;
  ioctl.in.flags = flags;
  ioctl.in.pBuffer = pBuffer;
  ioctl.in.length = length;
  ioctl.in.local = localAddress;
  return admxrc3IoctlNonBlock(hCard, pTicket, ADMXRC3_IOCTLCODE_READDMA, &ioctl, sizeof(ioctl.in), 0);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_StartReadDMALocked(
  ADMXRC3_HANDLE hCard,
  ADMXRC3_TICKET* pTicket,
  unsigned int dmaChannel,
  uint32_t flags,
  ADMXRC3_BUFFER_HANDLE hBuffer,
  size_t offset,
  size_t length,
  uint32_t localAddress)
{
  IOCTLS_ADMXRC3_READDMALOCKED ioctl;

  ioctl.in.channel = dmaChannel;
  ioctl.in.flags = flags;
  ioctl.in.hBuffer = hBuffer;
  ioctl.in.offset = offset;
  ioctl.in.length = length;
  ioctl.in.local = localAddress;
  return admxrc3IoctlNonBlock(hCard, pTicket, ADMXRC3_IOCTLCODE_READDMALOCKED, &ioctl, sizeof(ioctl.in), 0);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_StartReadDMALockedEx(
  ADMXRC3_HANDLE hCard,
  ADMXRC3_TICKET* pTicket,
  unsigned int dmaChannel,
  uint32_t flags,
  ADMXRC3_BUFFER_HANDLE hBuffer,
  size_t offset,
  size_t length,
  uint64_t localAddress)
{
  IOCTLS_ADMXRC3_READDMALOCKED ioctl;

  ioctl.in.channel = dmaChannel;
  ioctl.in.flags = flags;
  ioctl.in.hBuffer = hBuffer;
  ioctl.in.offset = offset;
  ioctl.in.length = length;
  ioctl.in.local = localAddress;
  return admxrc3IoctlNonBlock(hCard, pTicket, ADMXRC3_IOCTLCODE_READDMALOCKED, &ioctl, sizeof(ioctl.in), 0);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_ReadFlash(
  ADMXRC3_HANDLE hCard,
  unsigned int flashIndex,
  uint32_t flags,
  uint64_t offset,
  size_t length,
  void* pBuffer)
{
  IOCTLS_ADMXRC3_READFLASH ioctl;

  if (0 != length && NULL == pBuffer) {
    return ADMXRC3_NULL_POINTER;
  }

  ioctl.in.flashIndex = flashIndex;
  ioctl.in.flags = flags;
  ioctl.in.offset = offset;
  ioctl.in.length = length;
  ioctl.in.pBuffer = pBuffer;
  return admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_READFLASH, &ioctl, sizeof(ioctl.in), 0);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_ReadVPD(
  ADMXRC3_HANDLE hCard,
  uint32_t flags,
  size_t offset,
  size_t length,
  void* pBuffer)
{
  IOCTLS_ADMXRC3_READVPD ioctl;

  if (0 != length && NULL == pBuffer) {
    return ADMXRC3_NULL_POINTER;
  }

  ioctl.in.flags = flags;
  ioctl.in.offset = offset;
  ioctl.in.length = length;
  ioctl.in.pBuffer = pBuffer;
  return admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_READVPD, &ioctl, sizeof(ioctl.in), 0);
}

#if ADMXRC3_HAS_USER_EVENT_HANDLE

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
# if defined(_WIN32) || defined(_MSC_VER)
ADMXRC3_RegisterWin32Event(
# elif defined(__VXWORKS__) || defined(__vxworks)
ADMXRC3_RegisterVxwSem(
# else
# error "Can't detect OS."
# endif
  ADMXRC3_HANDLE hCard,
  _ADMXRC3_UINT32 notification,
  admxrc3_user_event_handle_t hEvent)
{
  IOCTLS_ADMXRC3_REGISTEREVENT ioctl;

  ioctl.in.notification = notification;
  ioctl.in.hEvent = hEvent;
  return admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_REGISTEREVENT, &ioctl, sizeof(ioctl.in), 0);
}
# endif

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_SetClockFrequency(
  ADMXRC3_HANDLE hCard,
  unsigned int clockIndex,
  uint32_t flags,
  double frequency,
  double* pActualFrequency)
{
  IOCTLS_ADMXRC3_SETCLOCKFREQUENCY ioctl;
  ADMXRC3_STATUS status;

  if (frequency > 9.0e18) {
    /*
    ** Exclude frequencies above 9.0e18 so that casts to int64_t (max. value 2^63-1)
    ** are logically safe (and 9.0e18 Hz is a ridiculous request unless you want to
    ** generate X-rays).
    */
    return ADMXRC3_INVALID_FREQUENCY;
  }

  ioctl.in.clockIndex = (uint32_t)clockIndex;
  ioctl.in.flags = flags;
  /* Casting to int64_t then uint64_t avoids missing '__fixunsdfdi' symbol when building as shared library with some GNU toolchains */
  ioctl.in.frequency = (uint64_t)(int64_t)frequency; /* Cast should be safe due to above range check */
  status = admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_SETCLOCKFREQUENCY, &ioctl, sizeof(ioctl.in), sizeof(ioctl.out));
  if (ADMXRC3_SUCCESS == status) {
    if (NULL != pActualFrequency) {
      /* Casting to int64_t then double avoids missing '__floadundidf' symbol when linking apps against shared library with some GNU toolchains */
      *pActualFrequency = (double)(int64_t)ioctl.out.actualFrequency; /* Cast should be safe due to above range check */
    }
  }
  return status;
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_StartNotificationWait(
  ADMXRC3_HANDLE hCard,
  ADMXRC3_TICKET* pTicket,
  _ADMXRC3_UINT32 notification)
{
  IOCTLS_ADMXRC3_NOTIFICATIONWAIT ioctl;

  ioctl.in.notification = notification;
  return admxrc3IoctlNonBlock(hCard, pTicket, ADMXRC3_IOCTLCODE_NOTIFICATIONWAIT, &ioctl, sizeof(ioctl.in), 0);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_StartWriteDMA(
  ADMXRC3_HANDLE hCard,
  ADMXRC3_TICKET* pTicket,
  unsigned int dmaChannel,
  uint32_t flags,
  const void* pBuffer,
  size_t length,
  uint32_t localAddress)
{
  IOCTLS_ADMXRC3_WRITEDMA ioctl;

  ioctl.in.channel = dmaChannel;
  ioctl.in.flags = flags;
  ioctl.in.pBuffer = pBuffer;
  ioctl.in.length = length;
  ioctl.in.local = localAddress;
  return admxrc3IoctlNonBlock(hCard, pTicket, ADMXRC3_IOCTLCODE_WRITEDMA, &ioctl, sizeof(ioctl.in), 0);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_StartWriteDMAEx(
  ADMXRC3_HANDLE hCard,
  ADMXRC3_TICKET* pTicket,
  unsigned int dmaChannel,
  uint32_t flags,
  const void* pBuffer,
  size_t length,
  uint64_t localAddress)
{
  IOCTLS_ADMXRC3_WRITEDMA ioctl;

  ioctl.in.channel = dmaChannel;
  ioctl.in.flags = flags;
  ioctl.in.pBuffer = pBuffer;
  ioctl.in.length = length;
  ioctl.in.local = localAddress;
  return admxrc3IoctlNonBlock(hCard, pTicket, ADMXRC3_IOCTLCODE_WRITEDMA, &ioctl, sizeof(ioctl.in), 0);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_StartWriteDMALocked(
  ADMXRC3_HANDLE hCard,
  ADMXRC3_TICKET* pTicket,
  unsigned int dmaChannel,
  uint32_t flags,
  ADMXRC3_BUFFER_HANDLE hBuffer,
  size_t offset,
  size_t length,
  uint32_t localAddress)
{
  IOCTLS_ADMXRC3_WRITEDMALOCKED ioctl;

  ioctl.in.channel = dmaChannel;
  ioctl.in.flags = flags;
  ioctl.in.hBuffer = hBuffer;
  ioctl.in.offset = offset;
  ioctl.in.length = length;
  ioctl.in.local = localAddress;
  return admxrc3IoctlNonBlock(hCard, pTicket, ADMXRC3_IOCTLCODE_WRITEDMALOCKED, &ioctl, sizeof(ioctl.in), 0);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_StartWriteDMALockedEx(
  ADMXRC3_HANDLE hCard,
  ADMXRC3_TICKET* pTicket,
  unsigned int dmaChannel,
  uint32_t flags,
  ADMXRC3_BUFFER_HANDLE hBuffer,
  size_t offset,
  size_t length,
  uint64_t localAddress)
{
  IOCTLS_ADMXRC3_WRITEDMALOCKED ioctl;

  ioctl.in.channel = dmaChannel;
  ioctl.in.flags = flags;
  ioctl.in.hBuffer = hBuffer;
  ioctl.in.offset = offset;
  ioctl.in.length = length;
  ioctl.in.local = localAddress;
  return admxrc3IoctlNonBlock(hCard, pTicket, ADMXRC3_IOCTLCODE_WRITEDMALOCKED, &ioctl, sizeof(ioctl.in), 0);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_SyncFlash(
  ADMXRC3_HANDLE hCard,
  unsigned int flashIndex,
  _ADMXRC3_UINT32 flags)
{
  IOCTLS_ADMXRC3_SYNCFLASH ioctl;

  ioctl.in.flashIndex = flashIndex;
  ioctl.in.flags = flags;
  return admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_SYNCFLASH, &ioctl, sizeof(ioctl.in), 0);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_Unconfigure(
  ADMXRC3_HANDLE hCard,
  unsigned int targetIndex,
  uint32_t flags)
{
  IOCTLS_ADMXRC3_UNCONFIGURE ioctl;
  ADMXRC3_STATUS status;

  ioctl.in.targetIndex = (uint32_t)targetIndex;
  ioctl.in.flags = flags;
  status = admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_UNCONFIGURE, &ioctl, sizeof(ioctl.in), 0);
  return status;
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_UnloadBitstreamA(
  ADMXRC3_BITSTREAMA* pBitstream)
{
  if (NULL == pBitstream) {
    return ADMXRC3_NULL_POINTER;
  }

  free(pBitstream);
  return ADMXRC3_SUCCESS;
}

#if PORTABLE_HAS_WCHAR_SUPPORT
ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_UnloadBitstreamW(
  ADMXRC3_BITSTREAMW* pBitstream)
{
  if (NULL == pBitstream) {
    return ADMXRC3_NULL_POINTER;
  }

  free(pBitstream);
  return ADMXRC3_SUCCESS;
}
#endif

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_Unlock(
  ADMXRC3_HANDLE hCard,
  ADMXRC3_BUFFER_HANDLE hBuffer)
{
  IOCTLS_ADMXRC3_UNLOCK ioctl;

  ioctl.in.hBuffer = hBuffer;
  return admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_UNLOCK, &ioctl, sizeof(ioctl.in), 0);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_UnmapWindow(
  ADMXRC3_HANDLE hCard,
  void* pVirtualBase)
{
  if (NULL == pVirtualBase) {
    return ADMXRC3_NULL_POINTER;
  }

  /* This has to be done in an OS-specific way */
  return admxrc3UnmapRegion(hCard, pVirtualBase);
}

#if ADMXRC3_HAS_USER_EVENT_HANDLE

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
# if defined(_WIN32) || defined(_MSC_VER)
ADMXRC3_UnregisterWin32Event(
# elif defined(__VXWORKS__) || defined(__vxworks)
ADMXRC3_UnregisterVxwSem(
# else
# error "Can't detect OS."
# endif
  ADMXRC3_HANDLE hCard,
  uint32_t notification,
  admxrc3_user_event_handle_t hEvent)
{
  IOCTLS_ADMXRC3_UNREGISTEREVENT ioctl;

  ioctl.in.notification = notification;
  ioctl.in.hEvent = hEvent;
  return admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_UNREGISTEREVENT, &ioctl, sizeof(ioctl.in), 0);
}
# endif

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_Write(
  ADMXRC3_HANDLE hCard,
  unsigned int windowIndex,
  uint32_t flags,
  size_t offset,
  size_t length,
  const void* pBuffer)
{
  IOCTLS_ADMXRC3_WRITE ioctl;

  ioctl.in.windowIndex = windowIndex;
  ioctl.in.flags = flags;
  ioctl.in.offset = offset;
  ioctl.in.length = length;
  ioctl.in.pBuffer = pBuffer;
  return admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_WRITE, &ioctl, sizeof(ioctl.in), 0);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_WriteDMA(
  ADMXRC3_HANDLE hCard,
  unsigned int dmaChannel,
  uint32_t flags,
  const void* pBuffer,
  size_t length,
  uint32_t localAddress)
{
  IOCTLS_ADMXRC3_WRITEDMA ioctl;

  ioctl.in.channel = dmaChannel;
  ioctl.in.flags = flags;
  ioctl.in.pBuffer = pBuffer;
  ioctl.in.length = length;
  ioctl.in.local = localAddress;
  return admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_WRITEDMA, &ioctl, sizeof(ioctl.in), 0);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_WriteDMAEx(
  ADMXRC3_HANDLE hCard,
  unsigned int dmaChannel,
  uint32_t flags,
  const void* pBuffer,
  size_t length,
  uint64_t localAddress)
{
  IOCTLS_ADMXRC3_WRITEDMA ioctl;

  ioctl.in.channel = dmaChannel;
  ioctl.in.flags = flags;
  ioctl.in.pBuffer = pBuffer;
  ioctl.in.length = length;
  ioctl.in.local = localAddress;
  return admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_WRITEDMA, &ioctl, sizeof(ioctl.in), 0);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_WriteDMALocked(
  ADMXRC3_HANDLE hCard,
  unsigned int dmaChannel,
  uint32_t flags,
  ADMXRC3_BUFFER_HANDLE hBuffer,
  size_t offset,
  size_t length,
  uint32_t localAddress)
{
  IOCTLS_ADMXRC3_WRITEDMALOCKED ioctl;

  ioctl.in.channel = dmaChannel;
  ioctl.in.flags = flags;
  ioctl.in.hBuffer = hBuffer;
  ioctl.in.offset = offset;
  ioctl.in.length = length;
  ioctl.in.local = localAddress;
  return admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_WRITEDMALOCKED, &ioctl, sizeof(ioctl.in), 0);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_WriteDMALockedEx(
  ADMXRC3_HANDLE hCard,
  unsigned int dmaChannel,
  uint32_t flags,
  ADMXRC3_BUFFER_HANDLE hBuffer,
  size_t offset,
  size_t length,
  uint64_t localAddress)
{
  IOCTLS_ADMXRC3_WRITEDMALOCKED ioctl;

  ioctl.in.channel = dmaChannel;
  ioctl.in.flags = flags;
  ioctl.in.hBuffer = hBuffer;
  ioctl.in.offset = offset;
  ioctl.in.length = length;
  ioctl.in.local = localAddress;
  return admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_WRITEDMALOCKED, &ioctl, sizeof(ioctl.in), 0);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_WriteFlash(
  ADMXRC3_HANDLE hCard,
  unsigned int flashIndex,
  uint32_t flags,
  uint64_t offset,
  size_t length,
  const void* pBuffer)
{
  IOCTLS_ADMXRC3_WRITEFLASH ioctl;

  if (0 != length && NULL == pBuffer) {
    return ADMXRC3_NULL_POINTER;
  }

  ioctl.in.flashIndex = flashIndex;
  ioctl.in.flags = flags;
  ioctl.in.offset = offset;
  ioctl.in.length = length;
  ioctl.in.pBuffer = pBuffer;
  return admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_WRITEFLASH, &ioctl, sizeof(ioctl.in), 0);
}

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_WriteVPD(
  ADMXRC3_HANDLE hCard,
  uint32_t flags,
  size_t offset,
  size_t length,
  const void* pBuffer)
{
  IOCTLS_ADMXRC3_WRITEVPD ioctl;

  if (0 != length && NULL == pBuffer) {
    return ADMXRC3_NULL_POINTER;
  }

  ioctl.in.flags = flags;
  ioctl.in.offset = offset;
  ioctl.in.length = length;
  ioctl.in.pBuffer = pBuffer;
  return admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_WRITEVPD, &ioctl, sizeof(ioctl.in), 0);
}

#if defined(_WIN32) || defined(_MSC_VER)
ADMXRC3_EXPORT void
ADMXRC3_CALLING_CONVENTION
ADMXRC3_SetWin32ExceptionLevel(
  unsigned int level)
{
#ifdef _DEBUG
  admxrc3SetExceptionLevel(level);
#endif
}
#endif
